﻿<%@ WebService Language="C#" Class="SPMetaData" %>

using System;
using System.Collections.Generic;
using System.Linq;
using System.Web;
using System.Web.Services;
using Microsoft.SharePoint;
using Microsoft.SharePoint.Utilities;
using System.Text;
using System.Configuration;
using System.Text.RegularExpressions;
using System.Xml;
using Microsoft.SharePoint.Taxonomy;


/// <summary>
/// Summary description for SPMetaData
/// </summary>
//[WebService(Namespace = "http://tempuri.org/")]
[WebServiceBinding(ConformsTo = WsiProfiles.BasicProfile1_1)]
// To allow this Web Service to be called from script, using ASP.NET AJAX, uncomment the following line. 
// [System.Web.Script.Services.ScriptService]
public class SPMetaData : System.Web.Services.WebService
{
    private bool runWithElevatedPrivileges = false;
    private static bool verbose = false;
    private static string logFilePath = string.Empty;
    private static bool fixClaimsUsername = false;
    private static int itemsPerPage;
    private static int pagesPerBatch;
    private static string currentPagingInfo = string.Empty;
    XmlDocument result = new XmlDocument();
    Dictionary<string, string> hash_TaxonomyFieldsMap = new Dictionary<string, string>();
    
    public SPMetaData () {

        //Uncomment the following line if using designed components 
        InitializeComponent(); 
    }

    #region WEBMETHODS

    //[WebMethod(Description = "This method gets a string of delimited SharePoint Lists URLs, a string of delimited metadata fields and a modification date in this format: mm/dd/yyyy hh:mm:ss AM or PM. It returns an xml of the SharePoint items IDs that were modified since the modification date with their field values.")]
    [WebMethod]
    public XmlDocument GetItemsFieldsDataByListURL(string lists, string fields, string lastModified, int nextItemID)
    {
        try
        {
            //StringBuilder res = new StringBuilder();
            DateTime _lastModified;

            if (!DateTime.TryParse(lastModified, out _lastModified))
            {
                _lastModified = new DateTime(2000, 1, 1);
            }


            if (runWithElevatedPrivileges)
            {
                SPSecurity.RunWithElevatedPrivileges(delegate()
                {
                    getDataByURLs(lists, fields, _lastModified, nextItemID, "list");
                });
            }
            else
            {
                getDataByURLs(lists, fields, _lastModified, nextItemID, "list");
            }

            return result;
        }
        catch (Exception ex)
        {
            result.LoadXml(ex.Message);
            return result;
        }
    }
    
    //[WebMethod(Description = "This method gets a string of delimited SharePoint Sites URLs, a string of delimited metadata fields and a modification date in this format: mm/dd/yyyy hh:mm:ss AM or PM. It returns an xml of the SharePoint items IDs that were modified since the modification date with their field values.")]
    [WebMethod]
    public XmlDocument GetItemsFieldsDataBySiteURL(string sites, string fields, string lastModified)
    {
        try
        {
            //StringBuilder res = new StringBuilder();
            DateTime _lastModified;

            if (!DateTime.TryParse(lastModified, out _lastModified))
            {
                _lastModified = new DateTime(2000, 1, 1);
            }


            if (runWithElevatedPrivileges)
            {
                SPSecurity.RunWithElevatedPrivileges(delegate()
                {
                    getDataByURLs(sites, fields, _lastModified, 0, "site");
                });
            }
            else
            {
                getDataByURLs(sites, fields, _lastModified, 0, "sites");
            }

            return result;
        }
        catch (Exception ex)
        {
            result.LoadXml(ex.Message);
            return result;
        }
    }
    
    #endregion WEBMETHODS

    #region METHODS

    private void InitializeComponent()
    {
        if (getWebConfigAppSettingsValue("SPMetadataRunWithElevatedPrivileges").ToLower() == "true")
        {
            runWithElevatedPrivileges = true;
        }

        if (getWebConfigAppSettingsValue("SPMetadataEnableLogging").ToLower() == "true")
        {
            verbose = true;
        }

        if (!string.IsNullOrEmpty(getWebConfigAppSettingsValue("SPMetadataLogFilePath")))
        {
            logFilePath = getWebConfigAppSettingsValue("SPMetadataLogFilePath");
        }

        if (getWebConfigAppSettingsValue("FixClaimsUserAccountRegex").ToLower() == "true")
        {
            fixClaimsUsername = true;
        }

        if (getWebConfigAppSettingsValue("SPMetadataItemsPerPage") != null)
        {
            if (!int.TryParse(getWebConfigAppSettingsValue("SPMetadataItemsPerPage"), out itemsPerPage))
            {
                itemsPerPage = 1000;
            }
        }

        if (getWebConfigAppSettingsValue("SPMetadataPagesPerBatch") != null)
        {
            if (!int.TryParse(getWebConfigAppSettingsValue("SPMetadataPagesPerBatch"), out pagesPerBatch))
            {
                pagesPerBatch = 10;
            }
        }

        currentPagingInfo = string.Empty;
    }

    private void getDataByURLs(string urls, string fields, DateTime lastModified, int nextItemID, string type)
    {
        string[] separators = { "|", "*" };
        string[] _urls = urls.Split(separators, StringSplitOptions.RemoveEmptyEntries);
        string innerXML = string.Empty;
        
        writeStartDocument();

        if (type == "list")
        {
            foreach (string list in _urls)
            {
                try
                {
                    innerXML = getItemsByListUrl(list, fields, lastModified, nextItemID);
                }
                catch
                { }

                writeListElement(list, currentPagingInfo, innerXML);
            }
        }
        else if (type == "site")
        {
            Dictionary<string, string> fields_vals = new Dictionary<string, string>();
            
            foreach (string site in _urls)
            {
                string siteSPID = string.Empty;
                
                try
                {
                    fields_vals = getSiteAttributes(site, fields, lastModified, out siteSPID);
                }
                catch
                { }

                writeSiteElement(siteSPID, fields_vals);
            }
        }
    }

    private Dictionary<string,string> getSiteAttributes(string siteUrl, string fields, DateTime lastModified, out string siteSPID)
    {
        string[] separators = { "|", "*" };
        List<string> _fields = new List<string>(fields.Split(separators, StringSplitOptions.RemoveEmptyEntries).Distinct());
        string result = string.Empty;
        Dictionary<string, string> fields_vals = new Dictionary<string, string>();
        List<string> nonExistingFields = new List<string>();
        
        siteSPID = string.Empty;
        
        using (SPSite _siteCollection = new SPSite(siteUrl))
        {
            using (SPWeb _SPWeb = _siteCollection.OpenWeb())
            {
                try
                {
                    siteSPID = _SPWeb.ID.ToString();
                    removeNonExistingFields(_SPWeb, _fields, nonExistingFields);
                    addManagedMetadataFields(_SPWeb, _fields, nonExistingFields);
                    
                    if (_fields.Count > 0)
                    {
                        writeLog("Metadata fields to load: " + String.Join(",", _fields.ToArray()));
                    }
                    else if (_fields.Count <= 0)
                    {
                        throw new Exception("No metadata fields to load");
                    }

                    foreach (string field in _fields)
                    {
                        string val = getSPWebProperty(_SPWeb, field);
                        fields_vals.Add(field, val);
                    }
                    
                    return fields_vals;
                }
                catch (ArgumentException ex1)
                {
                    writeLog("Error: Argument Exception (Field does not exist): " + ex1.Message + ". StackTrace: " + ex1.StackTrace);
                }
                catch (SPQueryThrottledException ex2)
                {
                    writeLog("Error: SPQueryThrottledException: " + ex2.Message);
                }
                catch (Exception ex3)
                {
                    writeLog("Error: " + ex3.Message);
                }
            }
        }

        return fields_vals;
    }

    private string getSPWebProperty(SPWeb web, string field)
    {
        if (web.GetProperty(field) != null && !string.IsNullOrEmpty(web.GetProperty(field).ToString()))
        {
            return web.GetProperty(field).ToString();
        }
        if (web.GetProperty(field.ToLower()) != null && !string.IsNullOrEmpty(web.GetProperty(field.ToLower()).ToString()))
        {
            return web.GetProperty(field.ToLower()).ToString();
        }

        return string.Empty;
    }
        
    private string getItemsByListUrl(string ListURL, string fields, DateTime lastModified, int nextItemID)
    {
        string[] separators = { "|", "*" };
        List<string> _fields = new List<string>(fields.Split(separators, StringSplitOptions.RemoveEmptyEntries).Distinct());
        List<string> nonExistingFields = new List<string>();
        string result = string.Empty;
        
        using (SPSite _siteCollection = new SPSite(ListURL))
        {
            using (SPWeb _SPWeb = _siteCollection.OpenWeb())
            {
                SPList list = _SPWeb.GetListFromUrl(ListURL);
                if (nextItemID == 0)
                {
                    writeLog("Iterating items in list: '" + ListURL + "'. Total items: " + list.ItemCount);
                }
                try
                {
                    removeNonExistingFields(list, _fields, nonExistingFields);
                    addManagedMetadataFields(list, _fields, nonExistingFields);
                    
                    if (_fields.Count > 0 && nextItemID == 0)
                    {
                        writeLog("Metadata fields to load: " + String.Join(",", _fields.ToArray()));
                    }
                    else if (_fields.Count <= 0)
                    {
                        throw new Exception("No metadata fields to load");
                    }
                    SPQuery oQuery = buildSPQueryForList(_fields, list, lastModified, nextItemID);
                    return getItemsByListWithPaging(list, oQuery, _fields, lastModified);

                }
                catch (ArgumentException ex1)
                {
                    writeLog("Error: Argument Exception (Field does not exist): " + ex1.Message + ". StackTrace: " + ex1.StackTrace);
                }
                catch (SPQueryThrottledException ex2)
                {
                    writeLog("Error: SPQueryThrottledException: " + ex2.Message);
                }
                catch (Exception ex3)
                {
                    writeLog("Error: " + ex3.Message);
                }
            }
        }

        return string.Empty;
    }
    
    private static string getWebConfigAppSettingsValue(string key)
    {
        string val = string.Empty;
        string tempVal = System.Configuration.ConfigurationManager.AppSettings[key];
        if (!string.IsNullOrEmpty(tempVal)) val = tempVal;
        return val;
    }

    private static SPQuery buildSPQueryForList(List<string> _fields, SPList oList, DateTime lastModified, int nextItemID)
    {
        StringBuilder fieldsQuery = new StringBuilder();
        SPQuery qry = new SPQuery();

        if (nextItemID > 0)
        {
            qry.ListItemCollectionPosition = new SPListItemCollectionPosition(@"Paged=TRUE&p_ID=" + nextItemID);
        }
        
        qry.Query =
        @"<Where><Geq><FieldRef Name='Modified' /><Value Type='DateTime'>" + SPUtility.CreateISO8601DateTimeFromSystemDateTime(lastModified) + @"</Value></Geq></Where><OrderBy><FieldRef Name='ID' /></OrderBy>";
        foreach (string field in _fields)
        {
            fieldsQuery.Append(string.Format("<FieldRef Name='{0}' />", field.Replace(" ", @"_x0020_")));
        }
        if (fieldsQuery.Length == 0)
        {
            throw new Exception("no fields to load");
        }
        qry.ViewFields = fieldsQuery.ToString();
        qry.RowLimit = (uint)itemsPerPage;
        
        return qry;
    }

    private string getItemsByListWithPaging(SPList oList, SPQuery oQuery, List<string> _fields, DateTime lastModified)
    {
        int currPageIndex = 1;
        int skippedItems = 0;
        StringBuilder res = new StringBuilder();
        
        do
        {
            writeLog("Current Page #" + currPageIndex);
            SPListItemCollection collListItems = oList.GetItems(oQuery);
            writeLog("Item(s) found: " + collListItems.Count);

            foreach (SPListItem oListItem in collListItems)
            {
                if (!isItemModified(oListItem, lastModified))
                {
                    ++skippedItems;
                    continue;
                }
                
                StringBuilder temp = new StringBuilder();
                Guid objectGuid = getSafeItemGuid(oList, oListItem);
                foreach (string field in _fields)
                {
                    string property = string.Empty;
                    if (TryGetProperty(oListItem, field, out property))
                    {
                        temp.Append("<" + resolveFieldName(field) + ">" + CleanInvalidXmlChars(escapeXMLValue(property)) + @"</" + resolveFieldName(field) + ">");
                    }
                }
                if (temp.Length > 0)
                {
                    res.Append(@"<item><spid>" + objectGuid.ToString() + @"</spid>" + temp + @"</item>");
                }
            }

            if (currPageIndex >= pagesPerBatch) // done with this batch, return
            {
                if (collListItems.Count < itemsPerPage || collListItems.ListItemCollectionPosition == null) //no more data available
                {
                    currentPagingInfo = string.Empty;
                }
                else // more data available
                {
                    // store current position
                    currentPagingInfo = collListItems.ListItemCollectionPosition.PagingInfo;
                }

                oQuery.ListItemCollectionPosition = null;
            }
            else
            {
                oQuery.ListItemCollectionPosition = collListItems.ListItemCollectionPosition;
            }
            
            currPageIndex++;
            
            if (skippedItems > 0)
            {
                writeLog("Item(s) skipped: " + skippedItems);
                skippedItems = 0;
            }
        } while (oQuery.ListItemCollectionPosition != null);

        return res.ToString();
    }
    
    private string resolveFieldName(string field)
    {
        var keyval = hash_TaxonomyFieldsMap.Where(m => m.Value.Equals(field, StringComparison.OrdinalIgnoreCase)).FirstOrDefault();
        if (keyval.Key != null)
        {
            return keyval.Key.Replace(" ", "_x0020_");
        }

        return field.Replace(" ", "_x0020_");
    }

    private static bool isItemModified(SPListItem oListItem, DateTime lastModified)
    {
        try
        {
            if (oListItem["Modified"] != null && DateTime.Parse(oListItem["Modified"].ToString()) > lastModified)
            {
                return true;
            }
            else
            {
                return false;
            }
        }
        catch(Exception ex)
        {
            writeLog("Could not get item last modification date. " + ex.Message);
            return true;
        }
    }
    
    private static bool TryGetProperty(SPListItem oListItem, string field, out string prop)
    {
        prop = string.Empty;

        try
        {

            try //try with space
            {
                if (oListItem[field] != null)
                {
                    prop = oListItem[field].ToString();
                }
            }
            catch (Exception ex)
            {
                if (ex.Message.Contains(@"Column '" + field + "' does not exist. It may have been deleted by another user."))
                {
                    field = field.Replace(" ", @"_x0020_");
                }
                else
                {
                    writeLog("Exception: " + ex.Message);
                }
            }
                
            if (oListItem[field] != null)
            {
                prop = oListItem[field].ToString();
            }
            int res;

            //username
            if (!string.IsNullOrEmpty(prop) && prop.Contains(";#") && int.TryParse(prop.Split(';')[0], out res) && !string.IsNullOrEmpty(prop.Split('#')[1]))
            {
                try
                {
                    string tempProp = resolveUserName(prop.Split('#')[1]);
                    if (!string.IsNullOrEmpty(tempProp))
                    {
                        prop = tempProp;
                    }
                }
                catch
                { }
            }
            
            if (!string.IsNullOrEmpty(prop))
            {
                return true;
            }
            else
            {
                return false;
            }
        }
        catch(Exception ex)
        {
            if (!ex.Message.Contains("Value does not fall within the expected range"))
            {
                writeLog("Could not get field '" + field + "' for item: '" + oListItem.Url + "'. " + ex.Message);
            }
            return false;
        }
    }
    
    private static Guid getSafeItemGuid(SPList oList, SPListItem oListItem)
    {
        return oList.GetItemById(oListItem.ID).UniqueId;
    }
    
    private static string escapeXMLValue(string value)
    {
        if (string.IsNullOrEmpty(value)) return value;
        string temp = value;
        temp = temp.Replace("&", "&amp;").Replace("<", "&lt;").Replace(">", "&gt;").Replace("\"", "&quot;").Replace("'", "&apos;");
        return temp;
    }

    private static string resolveUserName(string username)
    {
        //fix claims account name
        if(fixClaimsUsername)
        {
            Match res = Regex.Match(username, getWebConfigAppSettingsValue("FixClaimsUserAccountRegex"), RegexOptions.IgnoreCase | RegexOptions.IgnorePatternWhitespace);
            if (res.Success)
            {
                username = getUserFromRegex(res);
            }
        }

        return username;
    }

    private static string getUserFromRegex(Match matchResult)
    {
        if (string.IsNullOrEmpty(matchResult.Groups[2].Value))
        {
            return matchResult.Value;
        }
        else
        {
            return (matchResult.Groups[2].Value);
        }
    }

    private static void writeLog(string message)
    {
        if (verbose)
        {
            try
            {
                using (System.IO.StreamWriter sw = System.IO.File.AppendText(logFilePath.Replace(@"\", @"\\")))
                {
                    sw.Write(Environment.NewLine + DateTime.Now + "\t" + message);
                }
            }
            catch
            { }
        }
    }

    private void removeNonExistingFields(object obj, List<string> _fields, List<string> _nonExistingFields)
    {
        if (obj != null && _fields.Count > 0)
        {
            if (obj is SPList)
            {
                SPFieldCollection spFields = ((SPList)obj).Fields;
                
                foreach (string field in _fields)
                {
                    if (!spFields.ContainsField(field))
                    {
                        _nonExistingFields.Add(field);
                        writeLog("Could not find field '" + field + "' in '" + ((SPList)obj).Title + "', Ignoring.");
                    }
                }
            }
            else if (obj is SPWeb)
            {
                SPFieldCollection spFields = ((SPWeb)obj).Fields;
                
                foreach (string field in _fields)
                {
                    if (!(obj as SPWeb).AllProperties.ContainsKey(field) && !(obj as SPWeb).AllProperties.ContainsKey(field.ToLower()) && !spFields.ContainsField(field))
                    {
                        _nonExistingFields.Add(field);
                        writeLog("Could not find field '" + field + "' in '" + ((SPWeb)obj).Title + "', Ignoring.");
                    }
                }
            }

            _fields.RemoveAll(val => _nonExistingFields.Contains(val));
        }
    }

    /// <summary>
    /// This method adds field-termset mapping to hash_TaxonomyFieldsMap
    /// </summary>
    /// <param name="obj"></param>
    /// <param name="_fields"></param>
    /// <param name="nonExistingFields"></param>
    private void addManagedMetadataFields(object obj, List<string> _fields, List<string> nonExistingFields)
    {
        if (obj != null && nonExistingFields.Count > 0)
        {
            foreach (string termsetName in nonExistingFields)
            {
                if (hash_TaxonomyFieldsMap.ContainsKey(termsetName) && !_fields.Contains(termsetName))
                {
                    _fields.Add(hash_TaxonomyFieldsMap[termsetName]);
                }
                else
                {
                    string name;
                    if (tryGetFieldNamebyTermset(obj, termsetName, out name) && !string.IsNullOrEmpty(name))
                    {
                        hash_TaxonomyFieldsMap.Add(termsetName, name);
                        _fields.Add(name);
                    }
                }
            }
        }
    }

    /// <summary>
    /// This method tries to get the corresponding term-set name for a given field name
    /// </summary>
    /// <param name="obj"></param>
    /// <param name="field"></param>
    /// <param name="name"></param>
    /// <returns></returns>
    private bool tryGetFieldNamebyTermset(object obj, string termsetName, out string fieldName)
    {
        fieldName = string.Empty;
        
        try
        {
            Guid termsetID = Guid.Empty;
            TaxonomySession taxonomySession = null;
            
            if (obj is SPWeb)
            {
                taxonomySession = new TaxonomySession(((SPWeb)obj).Site, true);
            }
            else if (obj is SPList)
            {
                taxonomySession = new TaxonomySession(((SPList)obj).ParentWeb.Site, true);
            }
            else if (obj is SPSite)
            {
                taxonomySession = new TaxonomySession(((SPSite)obj), true);
            }
            else
            {
                fieldName = string.Empty;
                return false;
            }

            TermStore termstore = taxonomySession.DefaultSiteCollectionTermStore;
            TermSetCollection termsets = termstore.GetTermSets(termsetName, 1033);
            termsetID = termsets.Where(t => t.Name == termsetName).Select(t => t.Id).FirstOrDefault();

            if (termsetID != Guid.Empty)
            {
                foreach (SPField spfield in (obj is SPList ? ((SPList)obj).Fields : ((SPWeb)obj).Fields))
                {
                    if (spfield is TaxonomyField)
                    {
                        if ((spfield as TaxonomyField).TermSetId.Equals(termsetID))
                        {
                            fieldName = spfield.Title;
                        }
                    }
                }
            }
            
            return true;
        }
        catch
        {
            fieldName = string.Empty;
            return false;
        }
    }
    
    private static int resolveItemIdFromPagingInfo(string pagingInfo)
    {
        //Paged=TRUE&p_ID=20
        int nextItemID = -1;

        if (!int.TryParse(pagingInfo.Substring(pagingInfo.IndexOf("p_ID") + 5), out nextItemID))
        {
            nextItemID = -1;
        }

        return nextItemID;
    }

    private void writeStartDocument()
    {
        XmlNode docNode = result.CreateXmlDeclaration("1.0", "UTF-8", null);
        result.AppendChild(docNode);
        XmlNode treeNode = result.CreateElement("tree");
        result.AppendChild(treeNode);
    }


    private void writeSiteElement(string siteSPID, Dictionary<string,string> fields_vals)
    {
        XmlNode listNode = result.CreateElement("site");
        XmlAttribute urlAttribute = result.CreateAttribute("spid");
        urlAttribute.Value = siteSPID;
        listNode.Attributes.Append(urlAttribute);

        foreach (var key in fields_vals)
        {
            XmlAttribute fieldAttribute = result.CreateAttribute(key.Key.Replace(" ", "_x0020_"));
            fieldAttribute.Value = fields_vals[key.Key];
            listNode.Attributes.Append(fieldAttribute);
        }
        
        listNode.Attributes.RemoveNamedItem("xmlns");
        result.SelectSingleNode(@"/tree").AppendChild(listNode);
    }
      
    private void writeListElement(string list, string currentPagingInfo, string innerXML)
    {
        XmlNode listNode = result.CreateElement("list");
        XmlAttribute urlAttribute = result.CreateAttribute("url");
        urlAttribute.Value = list;
        listNode.Attributes.Append(urlAttribute);
        XmlAttribute hasMoreDataAttribute = result.CreateAttribute("hasmoredata");
        hasMoreDataAttribute.Value = (!string.IsNullOrEmpty(currentPagingInfo)) ? "true" : "false";
        listNode.Attributes.Append(hasMoreDataAttribute);
        if (!string.IsNullOrEmpty(currentPagingInfo))
        {
            XmlAttribute nextItemIdAttribute = result.CreateAttribute("nextitemid");
            nextItemIdAttribute.Value = resolveItemIdFromPagingInfo(currentPagingInfo).ToString();
            listNode.Attributes.Append(nextItemIdAttribute);
        }

        listNode.InnerXml = innerXML;
        listNode.Attributes.RemoveNamedItem("xmlns");
        result.SelectSingleNode(@"/tree").AppendChild(listNode);
    }

    private string CleanInvalidXmlChars(string xmlString)
    {
        string rgx = @"[^\x09\x0A\x0D\x20-\xD7FF\xE000-\xFFFD\x10000-x10FFFF]";
        return Regex.Replace(xmlString, rgx, "");
    }
    
    #endregion METHODS
}